@@ -1,36 +1,48 @@
-# -*- coding: utf-8 -*-
-#
 # Copyright (C) 2008 John Paulett (john -at- paulett.org)
-# Copyright (C) 2009, 2011, 2013 David Aguilar (davvid -at- gmail.com)
+# Copyright (C) 2009-2018 David Aguilar (davvid -at- gmail.com)
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
 # you should have received as part of this distribution.
-
-import sys
+from __future__ import absolute_import, division, unicode_literals
 import quopri
+import sys
 
-import jsonpickle.util as util
-import jsonpickle.tags as tags
-import jsonpickle.handlers as handlers
-
-from jsonpickle.compat import numeric_types, set, unicode
-from jsonpickle.backend import JSONBackend
+from . import util
+from . import tags
+from . import handlers
+from .compat import numeric_types, decodebytes
+from .backend import json
 
 
 def decode(string, backend=None, context=None, keys=False, reset=True,
-           safe=False):
-    backend = _make_backend(backend)
-    if context is None:
-        context = Unpickler(keys=keys, backend=backend, safe=safe)
-    return context.restore(backend.decode(string), reset=reset)
+           safe=False, classes=None):
+    """Convert a JSON string into a Python object.
+
+    The keyword argument 'keys' defaults to False.
+    If set to True then jsonpickle will decode non-string dictionary keys
+    into python objects via the jsonpickle protocol.
+
+    The keyword argument 'classes' defaults to None.
+    If set to a single class, or a sequence (list, set, tuple) of classes,
+    then the classes will be made available when constructing objects.  This
+    can be used to give jsonpickle access to local classes that are not
+    available through the global module import scope.
+
+    >>> decode('"my string"') == 'my string'
+    True
+    >>> decode('36')
+    36
+    """
+    backend = backend or json
+    context = context or Unpickler(keys=keys, backend=backend, safe=safe)
+    data = backend.decode(string)
+    return context.restore(data, reset=reset, classes=classes)
 
 
-def _make_backend(backend):
-    if backend is None:
-        return JSONBackend()
-    else:
-        return backend
+def _safe_hasattr(obj, attr):
+    """A safe (but slow) hasattr() that avoids hasattr"""
+    return attr in dir(obj)
 
 
 class _Proxy(object):
@@ -71,6 +83,15 @@ class _Proxy(object):
         self.instance = instance
 
 
+class _IDProxy(_Proxy):
+    def __init__(self, objs, index):
+        self._index = index
+        self._objs = objs
+
+    def get(self):
+        return self._objs[self._index]
+
+
 def _obj_setattr(obj, attr, proxy):
     setattr(obj, attr, proxy.get())
 
@@ -82,7 +103,7 @@ def _obj_setvalue(obj, idx, proxy):
 class Unpickler(object):
 
     def __init__(self, backend=None, keys=False, safe=False):
-        self.backend = _make_backend(backend)
+        self.backend = backend or json
         self.keys = keys
         self.safe = safe
 
@@ -102,25 +123,42 @@ class Unpickler(object):
         self._objs = []
         self._proxies = []
 
-    def restore(self, obj, reset=True):
+        # Extra local classes not accessible globally
+        self._classes = {}
+
+    def restore(self, obj, reset=True, classes=None):
         """Restores a flattened object to its original python state.
 
         Simply returns any of the basic builtin types
 
         >>> u = Unpickler()
-        >>> u.restore('hello world')
-        'hello world'
-        >>> u.restore({'key': 'value'})
-        {'key': 'value'}
+        >>> u.restore('hello world') == 'hello world'
+        True
+        >>> u.restore({'key': 'value'}) == {'key': 'value'}
+        True
 
         """
         if reset:
             self.reset()
+        if classes:
+            self.register_classes(classes)
         value = self._restore(obj)
         if reset:
             self._swap_proxies()
         return value
 
+    def register_classes(self, classes):
+        """Register one or more classes
+
+        :param classes: sequence of classes or a single class to register
+
+        """
+        if isinstance(classes, (list, tuple, set)):
+            for cls in classes:
+                self.register_classes(cls)
+        else:
+            self._classes[util.importable_name(classes)] = classes
+
     def _swap_proxies(self):
         """Replace proxies with their corresponding instances"""
         for (obj, attr, proxy, method) in self._proxies:
@@ -128,8 +166,10 @@ class Unpickler(object):
         self._proxies = []
 
     def _restore(self, obj):
-        if has_tag(obj, tags.BYTES):
-            restore = self._restore_bytestring
+        if has_tag(obj, tags.B64):
+            restore = self._restore_base64
+        elif has_tag(obj, tags.BYTES):  # Backwards compatibility
+            restore = self._restore_quopri
         elif has_tag(obj, tags.ID):
             restore = self._restore_id
         elif has_tag(obj, tags.REF):  # Backwards compatibility
@@ -155,10 +195,15 @@ class Unpickler(object):
         elif util.is_dictionary(obj):
             restore = self._restore_dict
         else:
-            restore = lambda x: x
+            def restore(x):
+                return x
         return restore(obj)
 
-    def _restore_bytestring(self, obj):
+    def _restore_base64(self, obj):
+        return decodebytes(obj[tags.B64].encode('utf-8'))
+
+    #: For backwards compatibility with bytes data produced by older versions
+    def _restore_quopri(self, obj):
         return quopri.decodestring(obj[tags.BYTES].encode('utf-8'))
 
     def _restore_iterator(self, obj):
@@ -170,12 +215,16 @@ class Unpickler(object):
         Assumes that iterator items (the last two) are represented as lists
         as per pickler implementation.
         """
+        proxy = _Proxy()
+        self._mkref(proxy)
         reduce_val = obj[tags.REDUCE]
         f, args, state, listitems, dictitems = map(self._restore, reduce_val)
 
-        if f == tags.NEWOBJ or f.__name__ == '__newobj__':
+        if f == tags.NEWOBJ or getattr(f, '__name__', '') == '__newobj__':
             # mandated special case
             cls = args[0]
+            if not isinstance(cls, type):
+                cls = self._restore(cls)
             stage1 = cls.__new__(cls, *args[1:])
         else:
             stage1 = f(*args)
@@ -186,11 +235,24 @@ class Unpickler(object):
             except AttributeError:
                 # it's fine - we'll try the prescribed default methods
                 try:
-                    stage1.__dict__.update(state)
+                    # we can't do a straight update here because we
+                    # need object identity of the state dict to be
+                    # preserved so that _swap_proxies works out
+                    for k, v in stage1.__dict__.items():
+                        state.setdefault(k, v)
+                    stage1.__dict__ = state
                 except AttributeError:
                     # next prescribed default
-                    for k, v in state.items():
-                        setattr(stage1, k, v)
+                    try:
+                        for k, v in state.items():
+                            setattr(stage1, k, v)
+                    except Exception:
+                        dict_state, slots_state = state
+                        if dict_state:
+                            stage1.__dict__.update(dict_state)
+                        if slots_state:
+                            for k, v in slots_state.items():
+                                setattr(stage1, k, v)
 
         if listitems:
             # should be lists if not None
@@ -204,17 +266,22 @@ class Unpickler(object):
             for k, v in dictitems:
                 stage1.__setitem__(k, v)
 
-        self._mkref(stage1)
+        proxy.reset(stage1)
+        self._swapref(proxy, stage1)
         return stage1
 
     def _restore_id(self, obj):
-        return self._objs[obj[tags.ID]]
+        try:
+            idx = obj[tags.ID]
+            return self._objs[idx]
+        except IndexError:
+            return _IDProxy(self._objs, idx)
 
     def _restore_ref(self, obj):
         return self._namedict.get(obj[tags.REF])
 
     def _restore_type(self, obj):
-        typeref = loadclass(obj[tags.TYPE])
+        typeref = loadclass(obj[tags.TYPE], classes=self._classes)
         if typeref is None:
             return obj
         return typeref
@@ -228,7 +295,7 @@ class Unpickler(object):
 
     def _restore_object(self, obj):
         class_name = obj[tags.OBJECT]
-        cls = loadclass(class_name)
+        cls = loadclass(class_name, classes=self._classes)
         handler = handlers.get(cls, handlers.get(class_name))
         if handler is not None:  # custom handler
             proxy = _Proxy()
@@ -244,7 +311,7 @@ class Unpickler(object):
         return self._restore_object_instance(obj, cls)
 
     def _restore_function(self, obj):
-        return loadclass(obj[tags.FUNCTION])
+        return loadclass(obj[tags.FUNCTION], classes=self._classes)
 
     def _loadfactory(self, obj):
         try:
@@ -267,16 +334,18 @@ class Unpickler(object):
         if has_tag(obj, tags.NEWARGSEX):
             args, kwargs = obj[tags.NEWARGSEX]
         else:
-            args = getargs(obj)
+            args = getargs(obj, classes=self._classes)
             kwargs = {}
         if args:
             args = self._restore(args)
         if kwargs:
             kwargs = self._restore(kwargs)
 
-        is_oldstyle = not (isinstance(cls, type) or getattr(cls, '__meta__', None))
+        is_oldstyle = (
+            not (isinstance(cls, type) or getattr(cls, '__meta__', None)))
         try:
-            if (not is_oldstyle) and hasattr(cls, '__new__'):  # new style classes
+            if (not is_oldstyle) and hasattr(cls, '__new__'):
+                # new style classes
                 if factory:
                     instance = cls.__new__(cls, factory, *args, **kwargs)
                     instance.default_factory = factory
@@ -293,7 +362,7 @@ class Unpickler(object):
             except TypeError:  # fail gracefully
                 try:
                     instance = make_blank_classic(cls)
-                except:  # fail gracefully
+                except Exception:  # fail gracefully
                     return self._mkref(obj)
 
         proxy.reset(instance)
@@ -302,7 +371,7 @@ class Unpickler(object):
         if isinstance(instance, tuple):
             return instance
 
-        if (hasattr(instance, 'default_factory') and
+        if (_safe_hasattr(instance, 'default_factory') and
                 isinstance(instance.default_factory, _Proxy)):
             instance.default_factory = instance.default_factory.get()
 
@@ -311,13 +380,14 @@ class Unpickler(object):
     def _restore_from_dict(self, obj, instance, ignorereserved=True):
         restore_key = self._restore_key_fn()
         method = _obj_setattr
+        deferred = {}
 
         for k, v in sorted(obj.items(), key=util.itemgetter):
             # ignore the reserved attribute
             if ignorereserved and k in tags.RESERVED:
                 continue
             if isinstance(k, numeric_types):
-                str_k = unicode(k)
+                str_k = k.__str__()
             else:
                 str_k = k
             self._namestack.append(str_k)
@@ -326,7 +396,17 @@ class Unpickler(object):
             value = self._restore(v)
             if (util.is_noncomplex(instance) or
                     util.is_dictionary_subclass(instance)):
-                instance[k] = value
+                try:
+                    if k == '__dict__':
+                        setattr(instance, k, value)
+                    else:
+                        instance[k] = value
+                except TypeError:
+                    # Immutable object, must be constructed in one shot
+                    if k != '__dict__':
+                        deferred[k] = value
+                    self._namestack.pop()
+                    continue
             else:
                 setattr(instance, k, value)
 
@@ -338,8 +418,14 @@ class Unpickler(object):
             # step out
             self._namestack.pop()
 
+        if deferred:
+            # SQLAlchemy Immutable mappings must be constructed in one shot
+            instance = instance.__class__(deferred)
+
+        return instance
+
     def _restore_object_instance_variables(self, obj, instance):
-        self._restore_from_dict(obj, instance)
+        instance = self._restore_from_dict(obj, instance)
 
         # Handle list and set subclasses
         if has_tag(obj, tags.SEQ):
@@ -366,12 +452,14 @@ class Unpickler(object):
             # implements described default handling
             # of state for object with instance dict
             # and no slots
-            self._restore_from_dict(state, instance, ignorereserved=False)
+            instance = self._restore_from_dict(
+                    state, instance, ignorereserved=False)
         elif has_slots:
-            self._restore_from_dict(state[1], instance, ignorereserved=False)
+            instance = self._restore_from_dict(
+                    state[1], instance, ignorereserved=False)
             if has_slots_and_dict:
-                self._restore_from_dict(state[0],
-                                        instance, ignorereserved=False)
+                instance = self._restore_from_dict(
+                        state[0], instance, ignorereserved=False)
         elif (not hasattr(instance, '__getnewargs__')
               and not hasattr(instance, '__getnewargs_ex__')):
             # __setstate__ is not implemented so that means that the best
@@ -387,9 +475,11 @@ class Unpickler(object):
         children = [self._restore(v) for v in obj]
         parent.extend(children)
         method = _obj_setvalue
-        proxies = [(parent, idx, value, method)
-                    for idx, value in enumerate(parent)
-                        if isinstance(value, _Proxy)]
+        proxies = [
+            (parent, idx, value, method)
+            for idx, value in enumerate(parent)
+            if isinstance(value, _Proxy)
+        ]
         self._proxies.extend(proxies)
         return parent
 
@@ -397,19 +487,23 @@ class Unpickler(object):
         return tuple([self._restore(v) for v in obj[tags.TUPLE]])
 
     def _restore_set(self, obj):
-        return set([self._restore(v) for v in obj[tags.SET]])
+        return {self._restore(v) for v in obj[tags.SET]}
 
     def _restore_dict(self, obj):
         data = {}
         restore_key = self._restore_key_fn()
         for k, v in sorted(obj.items(), key=util.itemgetter):
             if isinstance(k, numeric_types):
-                str_k = unicode(k)
+                str_k = k.__str__()
             else:
                 str_k = k
             self._namestack.append(str_k)
             k = restore_key(k)
             data[k] = self._restore(v)
+            # k is currently a proxy and must be replaced
+            if isinstance(data[k], _Proxy):
+                self._proxies.append((data, k, data[k], _obj_setvalue))
+
             self._namestack.pop()
         return data
 
@@ -428,7 +522,8 @@ class Unpickler(object):
         if self.keys:
             restore_key = self._restore_pickled_key
         else:
-            restore_key = lambda key: key
+            def restore_key(key):
+                return key
         return restore_key
 
     def _restore_pickled_key(self, key):
@@ -450,16 +545,14 @@ class Unpickler(object):
 
         >>> u = Unpickler()
         >>> u._namestack = []
-        >>> u._refname()
-        '/'
-
+        >>> u._refname() == '/'
+        True
         >>> u._namestack = ['a']
-        >>> u._refname()
-        '/a'
-
+        >>> u._refname() == '/a'
+        True
         >>> u._namestack = ['a', 'b']
-        >>> u._refname()
-        '/a/b'
+        >>> u._refname() == '/a/b'
+        True
 
         """
         return '/' + '/'.join(self._namestack)
@@ -488,7 +581,7 @@ class Unpickler(object):
         self._namedict[self._refname()] = instance
 
 
-def loadclass(module_and_name):
+def loadclass(module_and_name, classes=None):
     """Loads the module and returns the class.
 
     >>> cls = loadclass('datetime.datetime')
@@ -497,20 +590,35 @@ def loadclass(module_and_name):
 
     >>> loadclass('does.not.exist')
 
-    >>> loadclass('__builtin__.int')()
+    >>> loadclass('builtins.int')()
     0
 
     """
-    try:
-        module, name = module_and_name.rsplit('.', 1)
-        module = util.untranslate_module_name(module)
-        __import__(module)
-        return getattr(sys.modules[module], name)
-    except:
-        return None
+    # Check if the class exists in a caller-provided scope
+    if classes:
+        try:
+            return classes[module_and_name]
+        except KeyError:
+            pass
+    # Otherwise, load classes from globally-accessible imports
+    names = module_and_name.split('.')
+    # First assume that everything up to the last dot is the module name,
+    # then try other splits to handle classes that are defined within
+    # classes
+    for up_to in range(len(names)-1, 0, -1):
+        module = util.untranslate_module_name('.'.join(names[:up_to]))
+        try:
+            __import__(module)
+            obj = sys.modules[module]
+            for class_name in names[up_to:]:
+                obj = getattr(obj, class_name)
+            return obj
+        except (AttributeError, ImportError, ValueError) as ex:
+            continue
+    return None
 
 
-def getargs(obj):
+def getargs(obj, classes=None):
     """Return arguments suitable for __new__()"""
     # Let saved newargs take precedence over everything
     if has_tag(obj, tags.NEWARGSEX):
@@ -527,7 +635,7 @@ def getargs(obj):
         obj_dict = obj[tags.OBJECT]
     except KeyError:
         return []
-    typeref = loadclass(obj_dict)
+    typeref = loadclass(obj_dict, classes=classes)
     if not typeref:
         return []
     if hasattr(typeref, '_fields'):
